EXPORTED_SYMBOLS = ["httpsEnforcer"];

Ci = Components.interfaces;

Components.utils.import("resource://gre/modules/FileUtils.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");
Components.utils.import("resource://gre/modules/Services.jsm");

var httpsEnforcer = {
  _getWindow: function(id) {
    return Services.wm.getMostRecentWindow("navigator:browser");
  },

  init: function() {
    // add the button and show the installed message, if needed
    if (httpsEnforcer.prefs.isFirstRun()) {
      // put the button on the toolbar (far right) if not already there
      var navbar = httpsEnforcer._getWindow().document.getElementById("nav-bar");
      var curSet = navbar.currentSet;
      if (curSet.indexOf("httpsEnforcer-button") == -1) {
        var set = curSet + ',httpsEnforcer-button';
        navbar.setAttribute('currentset', set);
        navbar.currentSet = set;
        httpsEnforcer._getWindow().document.persist('nav-bar', 'currentset');
      }
      // give a quick one-time usage message
      Services.prompt.alert(null, "HTTPS Enforcer is now installed", "Click the lock button to disable, enable, and configure it.\n\nWhile red (enabled), unencrypted web requests will fail unless allowed explicitly.");
      httpsEnforcer.prefs.setFirstRun(false);
    }
    httpsEnforcer.button.updateAppearance();

    // load existing rules
    httpsEnforcer.rules.load();

    // start observing all http requests
    Services.obs.addObserver(httpsEnforcer, "http-on-modify-request", false);
  },

  observe: function(subject, topic, data) {
    if (topic == "http-on-modify-request" && httpsEnforcer.prefs.isEnabled()) {
      var request = subject.QueryInterface(Ci.nsIHttpChannel);
      if (!httpsEnforcer.rules.isAllowed(request.URI)) {
        if (httpsEnforcer.rules.isIgnored(request.URI)) {
          request.cancel(Components.results.NS_ERROR_ABORT);
        } else {
          var redirectUri = httpsEnforcer.rules.getRedirectUri(request.URI);
          if (redirectUri != null && httpsEnforcer.isRedirectSupported()) {
            request.redirectTo(redirectUri);
          } else {
            request.cancel(Components.results.NS_ERROR_ABORT);
            // signal that a block has occurred by briefly changing the badge
            var button = httpsEnforcer._getWindow().document.getElementById("httpsEnforcer-button");
            if (button != null) {
              if (button.getAttribute('status') != 'blocking') {
                button.setAttribute('status', 'blocking');
                httpsEnforcer._getWindow().setTimeout(function() {
                  httpsEnforcer.button.updateAppearance();
                }, 500);
              }
            }
            // update the recent list
            httpsEnforcer.recent.addURI(request.URI);
          }
        }
      }
    }
  },

  isRedirectSupported: function() {
    // redirect only supported with Gecko20/Firefox20
    var actualVersion = Services.appinfo.version;
    var requiredVersion = "20.0";
    return Services.vc.compare(actualVersion, requiredVersion) >= 0;
  },

  toggleEnabled: function() {
    httpsEnforcer.prefs.setEnabled(!httpsEnforcer.prefs.isEnabled());
    httpsEnforcer.button.updateAppearance();
  },

  clearRecent: function() {
    httpsEnforcer.recent.clear();
    httpsEnforcer.button.updateAppearance();
  },

  copyValueOf: function(element) {
    var value = element.getAttribute('value');
    const gClipboardHelper = Components.classes["@mozilla.org/widget/clipboardhelper;1"]
        .getService(Components.interfaces.nsIClipboardHelper);
    gClipboardHelper.copyString(value);
  }
}

httpsEnforcer.button = {

  updateAppearance: function(blocking) {
    var wins = Services.wm.getEnumerator(null);
    while (wins.hasMoreElements()) {
      var win = wins.getNext().QueryInterface(Components.interfaces.nsIDOMWindow);
      httpsEnforcer.button._updateAppearanceForWin(win, blocking);
    }
  },

  _updateAppearanceForWin: function(win, blocking) {
    var buttonElement = win.document.getElementById("httpsEnforcer-button");
    if (buttonElement != null) {
      if (blocking && button.getAttribute('status') != 'blocking') {
        button.setAttribute('status', 'blocking');
        httpsEnforcer._getWindow().setTimeout(function() {
          httpsEnforcer.button.updateAppearance();
        }, 250);
      } else if (httpsEnforcer.prefs.isEnabled()) {
        buttonElement.setAttribute('status', 'enabled');
        if (httpsEnforcer.prefs.getHideBadge() || httpsEnforcer.recent.blockCount == 0) {
          buttonElement.setAttribute('badgeLabel', '');
        } else {
          buttonElement.setAttribute('badgeLabel', httpsEnforcer.recent.blockCount);
        }
        buttonElement.tooltipText = "HTTPS Enforcer (Enabled)";
      } else {
        buttonElement.setAttribute('status', 'disabled');
        buttonElement.setAttribute('badgeLabel', '');
        buttonElement.tooltipText = "HTTPS Enforcer (Disabled)";
      }
    }
  },

  updateTopMenu: function() {
    // TODO: get urls from dtd
    var enabledImage = "chrome://https-enforcer/skin/httpsEnforcer-button-enabled.png";
    var disabledImage = "chrome://https-enforcer/skin/httpsEnforcer-button-disabled.png";
    var toggleEnabledElement = httpsEnforcer._getWindow().document.getElementById("httpsEnforcer-toggleEnabled");
    if (httpsEnforcer.prefs.isEnabled()) {
      toggleEnabledElement.image = disabledImage;
      toggleEnabledElement.label = "Disable HTTPS Enforcer";
    } else {
      toggleEnabledElement.image = enabledImage;
      toggleEnabledElement.label = "Enable HTTPS Enforcer";
    }

    var recentlyBlockedElement = httpsEnforcer._getWindow().document.getElementById("httpsEnforcer-recently-blocked");
    recentlyBlockedElement.label = "Recently Blocked (" + httpsEnforcer.recent.blockCount + ")";
  },

  updateRecentMenu: function() {
    var win = httpsEnforcer._getWindow();

    var recentlyBlockedPopup = win.document.getElementById("httpsEnforcer-recently-blocked-popup");
    if (recentlyBlockedPopup.state == 'open') return true; // no need to re-populate
    while (recentlyBlockedPopup.firstChild.tagName != "menuseparator") {
      recentlyBlockedPopup.removeChild(recentlyBlockedPopup.firstChild);
    }

    var orderedHostnames = httpsEnforcer.recent.getKeysOrderedByLastBlockedDate(httpsEnforcer.recent.hostInfo);
    orderedHostnames.reverse();
    for (var i = 0; i < orderedHostnames.length; i++) {
      var hostname = orderedHostnames[i];
      var hostInfo = httpsEnforcer.recent.hostInfo[hostname];
      var hostMenu = win.document.createElement("menu");
      hostMenu.setAttribute('label', 'From ' + hostname + ' (' + hostInfo.blockCount + ')');
      hostMenu.setAttribute('hostname', hostname);

      var hostMenuPopup = win.document.createElement("menupopup");
      hostMenuPopup.setAttribute('onpopupshowing', 'httpsEnforcer.button.updateRecentHostMenu(this);');
      hostMenu.appendChild(hostMenuPopup);

      recentlyBlockedPopup.insertBefore(hostMenu, recentlyBlockedPopup.firstChild);
    }

    return true;
  },

  updateRecentHostMenu: function(hostMenuPopup) {
    if (hostMenuPopup.state == 'open') return true; // no need to re-populate

    var win = httpsEnforcer._getWindow();

    while (hostMenuPopup.hasChildNodes()) {
      hostMenuPopup.removeChild(hostMenuPopup.firstChild);
    }

    var hostname = hostMenuPopup.parentNode.getAttribute('hostname');

    var allowImage = 'chrome://https-enforcer/skin/httpsEnforcer-allow.png';
    var ignoreImage = 'chrome://https-enforcer/skin/httpsEnforcer-ignore.png';
    var copyImage = 'chrome://https-enforcer/skin/httpsEnforcer-copy.png';

    httpsEnforcer.button._appendMenuItem(win, hostMenuPopup, "Allow all from this host", hostname, "httpsEnforcer.prefs.addAllow(document, window, this);", allowImage);
    httpsEnforcer.button._appendMenuItem(win, hostMenuPopup, "Ignore all from this host", hostname, "httpsEnforcer.prefs.addIgnore(document, window, this);", ignoreImage);
    httpsEnforcer.button._appendMenuItem(win, hostMenuPopup, "Copy this host", hostname, "httpsEnforcer.copyValueOf(this);", copyImage);

    hostMenuPopup.appendChild(win.document.createElement("menuseparator"));

    var hostInfo = httpsEnforcer.recent.hostInfo[hostname];
    var orderedUrls = httpsEnforcer.recent.getKeysOrderedByLastBlockedDate(hostInfo.urlInfo);
    for (var i = 0; i < orderedUrls.length; i++) {
      var url = orderedUrls[i];
      var urlInfo = hostInfo.urlInfo[url];
      var urlMenu = win.document.createElement("menu");
      urlMenu.setAttribute('label', url + ' (' + urlInfo.blockCount + ')');

      var urlMenuPopup = win.document.createElement("menupopup");
      httpsEnforcer.button._appendMenuItem(win, urlMenuPopup, "Allow this URL", url, "httpsEnforcer.prefs.addAllow(document, window, this);", allowImage);
      httpsEnforcer.button._appendMenuItem(win, urlMenuPopup, "Ignore this URL", url, "httpsEnforcer.prefs.addIgnore(document, window, this);", ignoreImage);
      httpsEnforcer.button._appendMenuItem(win, urlMenuPopup, "Copy this URL", url, "httpsEnforcer.copyValueOf(this);", copyImage);
      urlMenu.appendChild(urlMenuPopup);

      hostMenuPopup.appendChild(urlMenu);
    }

    return true;
  },

  _appendMenuItem: function(win, menupopup, label, value, oncommand, image) {
    var menuitem = win.document.createElement("menuitem");
    menuitem.setAttribute("label", label);
    menuitem.setAttribute("value", value);
    menuitem.setAttribute("oncommand", oncommand);
    if (image != null) {
      menuitem.setAttribute("class", "menuitem-iconic");
      menuitem.setAttribute("image", image);
    }
    menupopup.appendChild(menuitem);
    return menuitem;
  }
};

httpsEnforcer.prefs = {

  branch: Services.prefs.getBranch("extensions.httpsEnforcer."),

  isEnabled: function() {
    return httpsEnforcer.prefs.branch.getBoolPref("enabled");
  },

  setEnabled: function(value) {
    return httpsEnforcer.prefs.branch.setBoolPref("enabled", value);
  },

  isFirstRun: function() {
    return httpsEnforcer.prefs.branch.getBoolPref("firstRun");
  },

  setFirstRun: function(value) {
    httpsEnforcer.prefs.branch.setBoolPref("firstRun", value);
  },

  getMaxRecentlyBlockedHosts: function() {
    return httpsEnforcer.prefs.branch.getIntPref("maxRecentlyBlockedHosts");
  },

  setMaxRecentlyBlockedHosts: function(value) {
    httpsEnforcer.prefs.branch.setIntPref("maxRecentlyBlockedHosts", value);
  },

  getMaxRecentlyBlockedURLsPerHost: function() {
    return httpsEnforcer.prefs.branch.getIntPref("maxRecentlyBlockedURLsPerHost");
  },

  setMaxRecentlyBlockedURLsPerHost: function(value) {
    httpsEnforcer.prefs.branch.setIntPref("maxRecentlyBlockedURLsPerHost", value);
  },

  getAutoRedirect: function() {
    return httpsEnforcer.prefs.branch.getBoolPref("autoRedirect");
  },

  setAutoRedirect: function(value) {
    httpsEnforcer.prefs.branch.setBoolPref("autoRedirect", value);
  },

  getHideBadge: function() {
    return httpsEnforcer.prefs.branch.getBoolPref("hideBadge");
  },

  setHideBadge: function(value) {
    httpsEnforcer.prefs.branch.setBoolPref("hideBadge", value);
  },

  pageSelected: function(document, window) {
    var selectedItem = document.getElementById('httpsEnforcer-prefs-list').selectedItem;

    var iframe = document.getElementById('httpsEnforcer-prefs-iframe');
    iframe.setAttribute('src', selectedItem.value);

    var dialogheader = document.getElementById('httpsEnforcer-prefs-dialogheader');
    dialogheader.setAttribute('title', selectedItem.label);
  },

  dialogClosed: function(document, window) {
    httpsEnforcer.recent.refresh();
    httpsEnforcer.button.updateAppearance();
    return true;
  },

  generalPageLoaded: function(document, window) {
    document.getElementById('httpsEnforcer-prefs-maxRecentlyBlockedHosts').value = httpsEnforcer.prefs.getMaxRecentlyBlockedHosts();
    document.getElementById('httpsEnforcer-prefs-maxRecentlyBlockedURLsPerHost').value = httpsEnforcer.prefs.getMaxRecentlyBlockedURLsPerHost();
    if (httpsEnforcer.isRedirectSupported()) {
      document.getElementById('httpsEnforcer-prefs-autoRedirect').checked = httpsEnforcer.prefs.getAutoRedirect();
      document.getElementById('httpsEnforcer-prefs-hideBadge').checked = httpsEnforcer.prefs.getHideBadge();
    } else {
      document.getElementById('httpsEnforcer-prefs-autoRedirect').setAttribute('disabled', 'true');
      document.getElementById('httpsEnforcer-prefs-autoRedirectHeader').value = 'Auto-Redirect (Upgrade to Firefox 20.0+ to enable)';
      document.getElementById('httpsEnforcer-prefs-hideBadge').setAttribute('disabled', 'true');
      document.getElementById('httpsEnforcer-prefs-hideBadgeHeader').value = 'Hide Badge';
    }
    return true;
  },

  patternSelected: function(document, window) {
    document.getElementById('httpsEnforcer-prefs-editButton').setAttribute('disabled', 'false');
    document.getElementById('httpsEnforcer-prefs-deleteButton').setAttribute('disabled', 'false');
  },

  allowedPageLoaded: function(document, window) {
    httpsEnforcer.prefs._refreshList(
        document,
        document.getElementById('httpsEnforcer-prefs-allowed-listbox'),
        httpsEnforcer.rules.allowedPatterns);
    return true;
  },

  ignoredPageLoaded: function(document, window) {
    httpsEnforcer.prefs._refreshList(
        document,
        document.getElementById('httpsEnforcer-prefs-ignored-listbox'),
        httpsEnforcer.rules.ignoredPatterns);
    return true;
  },

  _promptForPattern: function(document, window, title, suggest) {
    var userpattern;
    if (!suggest) {
      suggest = "";
    }
    var input = { value: suggest };
    var text = 'Format is host/path or host:port/path\n\n(* anywhere matches any value)';
    var ok = Services.prompt.prompt(window, title, text, input, null, {value: false});
    if (!ok) {
      return null;
    }
    userpattern = input.value.trim();
    var pattern = userpattern;
    if (pattern.length == 0 || pattern === 'host:port/path') {
      return httpsEnforcer.prefs._promptForPattern(document, window, title, suggest);
    }
    if (pattern.indexOf("http://") != 0) {
      pattern = "http://" + pattern;
    }
    var j = pattern.indexOf('/');
    var afterScheme = pattern.substr(j + 2);
    j = afterScheme.indexOf('/');
    if (j < 0) {
      afterScheme = afterScheme + '/';
      j = afterScheme.indexOf('/');
    }
    // http:///
    var hostPort = afterScheme.substr(0, j).split(':');
    if (hostPort.length > 2) {
      window.alert("Too many colons in host:port");
      return httpsEnforcer.prefs._promptForPattern(document, window, title, userpattern);
    }
    var host = hostPort[0].trim();
    if (host.length == 0) {
      window.alert("No host specified");
      return httpsEnforcer.prefs._promptForPattern(document, window, title, userpattern);
    }
    var port = '80';
    if (hostPort.length == 2) {
      port = hostPort[1].trim();
      if (port.length == 0) {
        port = '80';
      }
    }
    var path = afterScheme.substr(j);

    return 'http://' + host + ":" + port + path;
  },

  addAllow: function(document, window, menuitem) {
    var suggest = menuitem.value;
    if (suggest.indexOf("http://") != 0) {
      suggest = suggest + ":*/*";
    } else {
      suggest = suggest.substr(7);
    }
    var pattern = httpsEnforcer.prefs._promptForPattern(
        document,
        window,
        'Add Allowed URL(s)',
        suggest);
    if (pattern != null) {
      httpsEnforcer.rules.allowedPatterns.push(pattern);
      httpsEnforcer.rules.save();
      httpsEnforcer.recent.refresh();
      httpsEnforcer.button.updateAppearance();
    }
  },

  addIgnore: function(document, window, menuitem) {
    var suggest = menuitem.value;
    if (suggest.indexOf("http://") != 0) {
      suggest = suggest + ":*/*";
    } else {
      suggest = suggest.substr(7);
    }
    var pattern = httpsEnforcer.prefs._promptForPattern(
        document,
        window,
        'Add Ignored URL(s)',
        suggest);
    if (pattern != null) {
      httpsEnforcer.rules.ignoredPatterns.push(pattern);
      httpsEnforcer.rules.save();
      httpsEnforcer.recent.refresh();
      httpsEnforcer.button.updateAppearance();
    }
  },

  addAllowed: function(document, window) {
    var pattern = httpsEnforcer.prefs._promptForPattern(
        document,
        window,
        'Add Allowed URL(s)',
        'host:port/path');
    if (pattern != null) {
      httpsEnforcer.rules.allowedPatterns.push(pattern);
      httpsEnforcer.rules.save();
      document.getElementById('httpsEnforcer-prefs-allowed-listbox').selectedIndex = -1;
      httpsEnforcer.prefs._refreshList(
          document,
          document.getElementById('httpsEnforcer-prefs-allowed-listbox'),
          httpsEnforcer.rules.allowedPatterns);
    }
  },

  editAllowed: function(document, window) {
    var oldPattern = document.getElementById('httpsEnforcer-prefs-allowed-listbox').selectedItem.value;
    var j = oldPattern.indexOf('/');
    var afterScheme = oldPattern.substr(j + 2);
    var pattern = httpsEnforcer.prefs._promptForPattern(
        document,
        window,
        'Edit Allowed URL(s)',
        afterScheme);
    if (pattern != null) {
      for (var i = 0; i < httpsEnforcer.rules.allowedPatterns.length; i++) {
        var value = httpsEnforcer.rules.allowedPatterns[i];
        if (value === oldPattern) {
          httpsEnforcer.rules.allowedPatterns[i] = pattern;
        }
      }
      httpsEnforcer.rules.save();
      document.getElementById('httpsEnforcer-prefs-allowed-listbox').selectedIndex = -1;
      httpsEnforcer.prefs._refreshList(
          document,
          document.getElementById('httpsEnforcer-prefs-allowed-listbox'),
          httpsEnforcer.rules.allowedPatterns);
    }
  },

  deleteAllowed: function(document, window) {
    var oldPattern = document.getElementById('httpsEnforcer-prefs-allowed-listbox').selectedItem.value;
    if (Services.prompt.confirm(window, "Delete Allowed URL?", oldPattern + "\n\nThis URL will no longer be allowed.")) {
      var newArray = new Array();
      for (var i = 0; i < httpsEnforcer.rules.allowedPatterns.length; i++) {
        var value = httpsEnforcer.rules.allowedPatterns[i];
        if (value != oldPattern) {
          newArray.push(value);
        }
      }
      httpsEnforcer.rules.allowedPatterns = newArray;
      httpsEnforcer.rules.save();
      document.getElementById('httpsEnforcer-prefs-allowed-listbox').selectedIndex = -1;
      httpsEnforcer.prefs._refreshList(
          document,
          document.getElementById('httpsEnforcer-prefs-allowed-listbox'),
          httpsEnforcer.rules.allowedPatterns);
    }
  },

  addIgnored: function(document, window) {
    var pattern = httpsEnforcer.prefs._promptForPattern(
        document,
        window,
        'Add Ignored URL',
        'host:port/path');
    if (pattern != null) {
      httpsEnforcer.rules.ignoredPatterns.push(pattern);
      httpsEnforcer.rules.save();
      document.getElementById('httpsEnforcer-prefs-ignored-listbox').selectedIndex = -1;
      httpsEnforcer.prefs._refreshList(
          document,
          document.getElementById('httpsEnforcer-prefs-ignored-listbox'),
          httpsEnforcer.rules.ignoredPatterns);
    }
  },

  editIgnored: function(document, window) {
    var oldPattern = document.getElementById('httpsEnforcer-prefs-ignored-listbox').selectedItem.value;
    var j = oldPattern.indexOf('/');
    var afterScheme = oldPattern.substr(j + 2);
    var pattern = httpsEnforcer.prefs._promptForPattern(
        document,
        window,
        'Edit Ignored URL',
        afterScheme);
    if (pattern != null) {
      for (var i = 0; i < httpsEnforcer.rules.ignoredPatterns.length; i++) {
        var value = httpsEnforcer.rules.ignoredPatterns[i];
        if (value === oldPattern) {
          httpsEnforcer.rules.ignoredPatterns[i] = pattern;
        }
      }
      httpsEnforcer.rules.save();
      document.getElementById('httpsEnforcer-prefs-ignored-listbox').selectedIndex = -1;
      httpsEnforcer.prefs._refreshList(
          document,
          document.getElementById('httpsEnforcer-prefs-ignored-listbox'),
          httpsEnforcer.rules.ignoredPatterns);
    }
  },

  deleteIgnored: function(document, window) {
    var oldPattern = document.getElementById('httpsEnforcer-prefs-ignored-listbox').selectedItem.value;
    if (Services.prompt.confirm(window, "Delete Ignored URL?", oldPattern + "\n\nThis URL will no longer be ignored.")) {
      var newArray = new Array();
      for (var i = 0; i < httpsEnforcer.rules.ignoredPatterns.length; i++) {
        var value = httpsEnforcer.rules.ignoredPatterns[i];
        if (value != oldPattern) {
          newArray.push(value);
        }
      }
      httpsEnforcer.rules.ignoredPatterns = newArray;
      httpsEnforcer.rules.save();
      document.getElementById('httpsEnforcer-prefs-ignored-listbox').selectedIndex = -1;
      httpsEnforcer.prefs._refreshList(
          document,
          document.getElementById('httpsEnforcer-prefs-ignored-listbox'),
          httpsEnforcer.rules.ignoredPatterns);
    }
  },

  _refreshList: function(document, listbox, patterns) {
    document.getElementById('httpsEnforcer-prefs-editButton').setAttribute('disabled', 'true');
    document.getElementById('httpsEnforcer-prefs-deleteButton').setAttribute('disabled', 'true');
    while (listbox.hasChildNodes()) {
      listbox.removeChild(listbox.firstChild);
    }

    var listhead = document.createElement('listhead');
    var hostHeader = document.createElement('listheader');
    hostHeader.setAttribute('label', 'Host');
    listhead.appendChild(hostHeader);
    var portHeader = document.createElement("listheader");
    portHeader.setAttribute('label', 'Port');
    listhead.appendChild(portHeader);
    var pathHeader = document.createElement("listheader");
    pathHeader.setAttribute('label', 'Path');
    listhead.appendChild(pathHeader);
    listbox.appendChild(listhead);

    var listcols = document.createElement('listcols');
    var listcol = document.createElement('listcol');
    listcols.appendChild(listcol);
    listcol = document.createElement('listcol');
    listcols.appendChild(listcol);
    listcol = document.createElement('listcol');
    listcol.setAttribute('flex', '1');
    listcols.appendChild(listcol);
    listbox.appendChild(listcols);

    patterns.sort(function(a, b){return a < b ? -1 : 1});

    for (var i = 0; i < patterns.length; i++) {
      var j = patterns[i].indexOf('/');
      var afterScheme = patterns[i].substr(j + 2);
      j = afterScheme.indexOf('/');
      var hostPort = afterScheme.substr(0, j).split(':');
      var host = hostPort[0];
      var port = hostPort[1];
      var path = afterScheme.substr(j);

      var listitem = document.createElement('listitem');
      listitem.setAttribute('value', patterns[i]);
      var hostCell = document.createElement('listcell');
      hostCell.setAttribute('label', host + ' ');
      listitem.appendChild(hostCell);
      var portCell = document.createElement('listcell');
      portCell.setAttribute('label',  ' ' + port + ' ');
      listitem.appendChild(portCell);
      var pathCell = document.createElement('listcell');
      pathCell.setAttribute('label', ' ' + path);
      listitem.appendChild(pathCell);
      listbox.appendChild(listitem);
    }
  }
};

httpsEnforcer.recent = {

  hostInfo: {},

  blockCount: 0,

  clear: function() {
    httpsEnforcer.recent.hostInfo = {};
    httpsEnforcer.recent.blockCount = 0;
  },

  addURI: function(uri) {
    var now = new Date().getTime();

    // update hostInfo as appropriate
    var hostInfo = httpsEnforcer.recent.hostInfo[uri.host];
    if (hostInfo == null) {
      // add new host to the recent hostInfo list
      hostInfo = {
        blockCount: 0,
        urlInfo: {}
      };
      httpsEnforcer.recent.hostInfo[uri.host] = hostInfo;
    }
    hostInfo.blockCount += 1;
    hostInfo.lastBlockedDate = now;
    httpsEnforcer.recent.dropOldHostInfo();

    // update hostInfo.urlInfo as appropriate
    var urlInfo = hostInfo.urlInfo[uri.spec];
    if (urlInfo == null) {
      urlInfo = {
        blockCount: 0
      };
      hostInfo.urlInfo[uri.spec] = urlInfo;
    }
    urlInfo.blockCount += 1;
    urlInfo.lastBlockedDate = now;
    httpsEnforcer.recent.dropOldUrlInfo(hostInfo);

    httpsEnforcer.recent.recalculateBlockCount();
  },

  dropOldHostInfo: function() {
    while (Object.keys(httpsEnforcer.recent.hostInfo).length > httpsEnforcer.prefs.getMaxRecentlyBlockedHosts()) {
      // delete the hostInfo with the oldest blocked date
      var orderedHostnames = httpsEnforcer.recent.getKeysOrderedByLastBlockedDate(httpsEnforcer.recent.hostInfo);
      delete httpsEnforcer.recent.hostInfo[orderedHostnames[orderedHostnames.length - 1]];
    }
  },

  dropOldUrlInfo: function(hostInfo) {
    while (Object.keys(hostInfo.urlInfo).length > httpsEnforcer.prefs.getMaxRecentlyBlockedURLsPerHost()) {
      // delete the urlInfo with the oldest blocked date
      var orderedUrls = httpsEnforcer.recent.getKeysOrderedByLastBlockedDate(hostInfo.urlInfo);
      var oldUrlInfo = hostInfo.urlInfo[orderedUrls[orderedUrls.length - 1]];
      delete hostInfo.urlInfo[orderedUrls[orderedUrls.length - 1]];
      // recalculate the block count for this hostInfo
      hostInfo.blockCount -= oldUrlInfo.blockCount;
    }
  },

  recalculateBlockCount: function() {
    var blockCount = 0;
    for (var hostname in httpsEnforcer.recent.hostInfo) {
      blockCount += httpsEnforcer.recent.hostInfo[hostname].blockCount;
    }
    httpsEnforcer.recent.blockCount = blockCount;
  },

  refresh: function() {
    // remove all urls that match any allowed or ignored urls
    for (var hostname in httpsEnforcer.recent.hostInfo) {
      var hostInfo = httpsEnforcer.recent.hostInfo[hostname];
      for (var url in hostInfo.urlInfo) {
        // remove this urlInfo from this hostInfo if it's no longer blocked
        var uri = Services.io.newURI(url, null, null);
        if (httpsEnforcer.rules.isAllowed(uri) || httpsEnforcer.rules.isIgnored(uri)) {
          var oldUrlInfo = hostInfo.urlInfo[url];
          delete hostInfo.urlInfo[url];
          hostInfo.blockCount -= oldUrlInfo.blockCount;
        }
      }
      if (Object.keys(hostInfo.urlInfo).length == 0) {
        // remove hostInfo if no more urlInfo for it
        delete httpsEnforcer.recent.hostInfo[hostname];
      }
      httpsEnforcer.recent.recalculateBlockCount();
    }

    // remove oldest urlinfo and hostinfo if needed to fit within maximums
    httpsEnforcer.recent.dropOldHostInfo();
    for (var hostname in httpsEnforcer.recent.hostInfo) {
      httpsEnforcer.recent.dropOldUrlInfo(httpsEnforcer.recent.hostInfo[hostname]);
    }
    httpsEnforcer.recent.recalculateBlockCount();
  },

  getKeysOrderedByLastBlockedDate: function(info) {
    return Object.keys(info).sort(function(a, b) {
      return info[b].lastBlockedDate - info[a].lastBlockedDate;
    });
  }
};

httpsEnforcer.rules = {

  _recentRedirectUris: { },

  filename: "httpsEnforcer-rules.json",

  allowedPatterns: [],

  ignoredPatterns: [],

  load: function() {
    var file = FileUtils.getFile("ProfD", [httpsEnforcer.rules.filename]);
    if (file.exists()) {
      NetUtil.asyncFetch(file, function(inputStream, status) {
        if (!Components.isSuccessCode(status)) {
          Services.console.logStringMessage("ERROR: Unable to load " + httpsEnforcer.rules.filename + " (status code " + status + ")");
        } else {
          var json = NetUtil.readInputStreamToString(inputStream, inputStream.available());
          var rules = JSON.parse(json);
          httpsEnforcer.rules.allowedPatterns = rules.allowedPatterns;
          httpsEnforcer.rules.ignoredPatterns = rules.ignoredPatterns;
        }
      });
    }
  },

  save: function() {
    var json = JSON.stringify({
      allowedPatterns: httpsEnforcer.rules.allowedPatterns,
      ignoredPatterns: httpsEnforcer.rules.ignoredPatterns
    });

    var file = FileUtils.getFile("ProfD", [httpsEnforcer.rules.filename]);
    var ostream = FileUtils.openSafeFileOutputStream(file);
    var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
        createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
    converter.charset = "UTF-8";
    var istream = converter.convertToInputStream(json);

    NetUtil.asyncCopy(istream, ostream, function(status) {
      if (!Components.isSuccessCode(status)) {
        Services.console.logStringMessage("ERROR: Unable to save " + httpsEnforcer.rules.filename + " (status code " + status + ")");
      }
    });
  },

  isAllowed: function(uri) {
    return uri.scheme == 'https' || httpsEnforcer.rules._matchesAny(uri, httpsEnforcer.rules.allowedPatterns);
  },

  isIgnored: function(uri) {
    return httpsEnforcer.rules._matchesAny(uri, httpsEnforcer.rules.ignoredPatterns);
  },

  getRedirectUri: function(uri) {
    if (uri.scheme == 'https') return null;

    var barValue = httpsEnforcer._getWindow().gURLBar.value;
    var remainder = uri.host + uri.path;

    // never redirect OCSP urls
    if (remainder.indexOf("ocsp") != -1) return null;

    if (httpsEnforcer.prefs.getAutoRedirect() || remainder.indexOf(barValue) == 0) {
      var redirectUri = Services.io.newURI('https://' + remainder, null, null);

      // detect redirect loop; if same url has already been redirected
      // twice in the last four seconds, don't redirect again

      var now = new Date().getTime();
      var before = now - 4000;

      // remove old dates first
      for (var u in httpsEnforcer.rules._recentRedirectUris) {
        var dates = httpsEnforcer.rules._recentRedirectUris[u];
        while (dates.length > 0 && dates[0] < before) {
          dates.shift();
        }
        if (dates.length == 0) {
          delete httpsEnforcer.rules._recentRedirectUris[u];
        }
      }

      // add this one
      var dates = httpsEnforcer.rules._recentRedirectUris[remainder];
      if (dates == null) {
        dates = new Array();
        httpsEnforcer.rules._recentRedirectUris[remainder] = dates;
      }
      dates.push(now);

      // if this was the second, block it
      if (dates.length > 2) {
        return null;
      }

      return redirectUri;
    }
    return null;
  },

  _matchesAny: function(uri, patterns) {
    for (var i = 0; i < patterns.length; i++) {
      if (httpsEnforcer.rules._matches(uri, patterns[i])) {
        return true;
      }
    }
    return false;
  },

  _matches: function(uri, pattern) {
    var portNum = uri.port;
    if (portNum == -1) {
      portNum = 80;
    }
    var normalizedUri = uri.scheme + '://' + uri.host + ':' + portNum + uri.path;

    var regExp;
    if (pattern.indexOf('~') == 0) {
      regExp = new RegExp(pattern);
    } else {
      regExp = httpsEnforcer.rules._patternToRegExp(pattern);
    }

    return regExp.test(normalizedUri);
  },

  _patternToRegExp: function(pattern) {
    var re = '^' + pattern.replace(/[-\/\\^$+?.()|[\]{}]/g, '\\$&').replace(/\*+/g, '.*') + '$';
    return new RegExp(re);
  }
};
